def plot_linear_trend(loader, var_name, date, trend_leadtime):
    """Plot linear trend against source data

    This is a method to plot linear trends against the corresponding variable
    data from a given preprocessing run.

    :return:
    """
